"""Tests for basic functionality."""

from pathlib import Path

import pydantic
import pytest
from pydantic import BaseModel

from pydantic_yaml import parse_yaml_file_as, parse_yaml_raw_as, to_yaml_str, to_yaml_file

from pydantic_yaml.examples.base_models import (
    A,
    B,
    CustomRootListObj,
    CustomRootListStr,
    Empty,
    HasEnums,
    SecretTstModel,
    SecretTstModelDumpable,
    UsesRefs,
    root,
)


@pytest.mark.parametrize(
    ["model_type", "fn"],
    [
        (Empty, "a.yaml"),
        (A, "a.yaml"),
        (A, "a-1.1.yaml"),
        (A, "a-1.2.yaml"),
        (B, "b.yaml"),
        (UsesRefs, "uses_refs.yaml"),
        (HasEnums, "has_enums.yaml"),
        (CustomRootListStr, "root_list_str.yaml"),
        (CustomRootListObj, "root_list_obj.yaml"),
    ],
)
def test_load_rt_simple_files(model_type: type[BaseModel], fn: str):
    """Test simple file loading and roundtripping."""
    # Load file
    file = root / fn
    obj = parse_yaml_file_as(model_type, file)
    # Roundtrip
    raw = to_yaml_str(obj)
    obj_rt = parse_yaml_raw_as(model_type, raw)
    # Check equality
    assert obj_rt == obj


@pytest.mark.parametrize("model", [A(a="aaa")])
def test_write_simple_model(model: BaseModel):
    """Test output of simple models."""
    to_yaml_str(model)  # TODO: Check output vs expected?


@pytest.mark.xfail(pydantic.VERSION >= "2", reason="Pydantic v2 is stricter for Bytes types.")
def test_secret_no_rt():
    """Test secret models properly failing to roundtrip."""
    sm = SecretTstModel(ss="123", sb=b"321")  # type: ignore
    assert sm.ss.get_secret_value() == "123"
    assert sm.sb.get_secret_value() == b"321"

    raw = to_yaml_str(sm)
    mdl = parse_yaml_raw_as(SecretTstModel, raw)
    assert mdl.ss.get_secret_value() != "123"
    assert mdl.sb.get_secret_value() != b"321"


@pytest.mark.xfail(pydantic.VERSION >= "2", reason="Pydantic v2 is stricter for Bytes types.")
def test_secret_yes_rt():
    """Test 'fixed' secret models properly roundtripping."""
    sm = SecretTstModelDumpable(ss="123", sb=b"321")  # type: ignore
    assert sm.ss.get_secret_value() == "123"
    assert sm.sb.get_secret_value() == b"321"

    raw = to_yaml_str(sm)
    mdl = parse_yaml_raw_as(SecretTstModelDumpable, raw)
    assert mdl.ss.get_secret_value() == "123"
    assert mdl.sb.get_secret_value() == b"321"


def test_write_open_file(tmpdir):
    """Test writing to a pre-opened file."""
    with (Path(tmpdir) / "test_write_open_file.yaml").open(mode="w") as f:
        to_yaml_file(f, A(a="a"))



"""Models used for testing, just normal `pydantic.BaseModel` objects."""

__all__ = [
    "Empty",
    "A",
    "B",
    "Recursive",
    "SecretTstModel",
    "SecretTstModelDumpable",
    "MyStrEnum",
    "MyIntEnum",
    "HasEnums",
    "UsesRefs",
    "CustomRootListStr",
    "CustomRootListObj",
    "root",
]

from pathlib import Path

import pydantic

from .common import MyIntEnum, MyStrEnum

root = Path(__file__).resolve().parent / "base_models"

if pydantic.version.VERSION < "2":
    from .v1.base_models import (
        A,
        B,
        CustomRootListObj,
        CustomRootListStr,
        Empty,
        HasEnums,
        Recursive,
        SecretTstModel,
        SecretTstModelDumpable,
        UsesRefs,
    )
else:
    from .v2.base_models import (  # type: ignore
        A,
        B,
        CustomRootListObj,
        CustomRootListStr,
        Empty,
        HasEnums,
        Recursive,
        SecretTstModel,
        SecretTstModelDumpable,
        UsesRefs,
    )



"""Test custom dumping behavior."""

import pytest
from pydantic import BaseModel

from pydantic_yaml import to_yaml_str
from pydantic_yaml.examples.base_models import HasEnums
from pydantic_yaml.examples.common import MyIntEnum, MyStrEnum

has_enums = HasEnums(opts=MyStrEnum.option1, vals=[MyIntEnum.v1, MyIntEnum.v2])


@pytest.mark.parametrize(
    ["mdl", "kwargs", "expected"],
    [
        [has_enums, dict(default_flow_style=False), "opts: option1\nvals:\n- 1\n- 2\n"],
        [has_enums, dict(default_flow_style=True), "{opts: option1, vals: [1, 2]}\n"],
        [has_enums, dict(indent=4), "opts: option1\nvals:\n    - 1\n    - 2\n"],
        [has_enums, dict(indent=6, map_indent=2), "opts: option1\nvals:\n      - 1\n      - 2\n"],
    ],
)
def test_dump_kwargs(mdl: BaseModel, kwargs: dict, expected: str):
    """Test dumping keyword arguments."""
    res = to_yaml_str(mdl, **kwargs)
    assert res == expected



"""Models used for testing, just normal `pydantic.BaseModel` objects."""

# mypy: ignore-errors

from typing import Optional

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    field_serializer,
    model_serializer,
    model_validator,
)
from pydantic.types import SecretBytes, SecretStr

from pydantic_yaml.examples.common import MyIntEnum, MyStrEnum


class Empty(BaseModel):
    """Empty model."""


class A(BaseModel):
    """Class A."""

    a: str


class B(BaseModel):
    """Class B."""

    b: str


class Recursive(BaseModel):
    """Recursive model, which is actually unsupported."""

    inner: Optional["Recursive"]
    a: int  # Doesn't work!


Recursive.model_rebuild()


class SecretTstModel(BaseModel):
    """Normal model with secret fields. This can't be roundtripped."""

    ss: SecretStr
    sb: SecretBytes


def _encode_secret(obj: SecretStr | SecretBytes | None) -> str | bytes | None:
    """Encode secret value."""
    if obj is None:
        return None
    return obj.get_secret_value()


class SecretTstModelDumpable(SecretTstModel):
    """Round-trippable model. This will save secret fields as the raw values."""

    @field_serializer("ss", "sb")
    def _serialize_secrets(self, secret: SecretStr | SecretBytes, _info):
        """Serialize secret fields.

        See Also
        --------
        https://docs.pydantic.dev/latest/usage/serialization/#custom-serializers
        """
        return _encode_secret(secret)


class HasEnums(BaseModel):
    """Base model with enums."""

    opts: MyStrEnum
    vals: list[MyIntEnum]


class _Name(BaseModel):  # type: ignore[no-redef]
    """First/last names."""

    given: str
    family: str


class UsesRefs(BaseModel):  # type: ignore[no-redef]
    """Example for the reference data."""

    bill_to: _Name = Field(alias="bill-to")
    ship_to: _Name = Field(alias="ship-to")

    # Pydantic config
    model_config = ConfigDict(populate_by_name=True)


class CustomRootListStr(BaseModel):  # type: ignore[no-redef]
    """Model with a custom root type.

    See Also
    --------
    https://docs.pydantic.dev/blog/pydantic-v2-alpha/#changes-to-basemodel
    https://github.com/pydantic/pydantic/blob/2b9459f20d094a46fa3093b43c34444240f03646/tests/test_parse.py#L95-L113
    """

    root: list[str]

    @model_validator(mode="before")
    @classmethod
    def _populate_root(cls, values):
        if isinstance(values, CustomRootListStr):
            return values
        return {"root": values}

    @model_serializer(mode="wrap")
    def _serialize(self, handler, info):
        data = handler(self)
        if info.mode == "json":
            return data["root"]
        else:
            return data

    @classmethod
    def model_modify_json_schema(cls, json_schema):
        """JSON schema changer."""
        return json_schema["properties"]["root"]


class CustomRootListObj(BaseModel):  # type: ignore[no-redef]
    """Model with a custom root type, list of objects.

    See Also
    --------
    https://docs.pydantic.dev/blog/pydantic-v2-alpha/#changes-to-basemodel
    https://github.com/pydantic/pydantic/blob/2b9459f20d094a46fa3093b43c34444240f03646/tests/test_parse.py#L95-L113
    """

    root: list[A | B]

    @model_validator(mode="before")
    @classmethod
    def _populate_root(cls, values):
        if isinstance(values, CustomRootListObj):
            return values
        return {"root": values}

    @model_serializer(mode="wrap")
    def _serialize(self, handler, info):
        data = handler(self)
        if info.mode == "json":
            return data["root"]
        else:
            return data

    @classmethod
    def model_modify_json_schema(cls, json_schema):
        """JSON schema changer."""
        return json_schema["properties"]["root"]



"""YAML-enabled Pydantic models."""

__all__ = [
    # New API
    "__version__",
    "parse_yaml_file_as",
    "parse_yaml_raw_as",
    "to_yaml_file",
    "to_yaml_str",
    # TODO: Re-add
    # YamlModelMixin ?
    # YamlModel ?
    # Deprecated classes
    "YamlInt",
    "YamlIntEnum",
    "YamlStr",
    "YamlStrEnum",
]


from pydantic.version import VERSION as PYDANTIC_VERSION

if (PYDANTIC_VERSION > "1") and (PYDANTIC_VERSION < "2"):
    from pydantic_yaml._internals.v1 import (
        to_yaml_file,
        to_yaml_str,
        parse_yaml_file_as,
        parse_yaml_raw_as,
    )
elif (PYDANTIC_VERSION > "2") and (PYDANTIC_VERSION < "3"):
    from pydantic_yaml._internals.v2 import (
        to_yaml_file,
        to_yaml_str,
        parse_yaml_file_as,
        parse_yaml_raw_as,
    )
else:
    raise ImportError(f"Unsupported version of Pydantic: {PYDANTIC_VERSION!r}")

from .deprecated.types import YamlInt, YamlIntEnum, YamlStr, YamlStrEnum
from .version import __version__



"""Tests for the basic functionality advertised in README."""


def test_import():
    """Ensure pydantic_yaml can be imported."""
    from pydantic_yaml import __version__

    assert __version__ != "0.0.0"



"""Test for dataclasses in Pydantic v2."""

import pytest

from pydantic.version import VERSION as PYDANTIC_VERSION
from pydantic.dataclasses import dataclass
from pydantic_yaml import to_yaml_str


@dataclass
class ExampleDataclass:
    """Example dataclass."""

    foo: str = "bar"


@pytest.mark.skipif(PYDANTIC_VERSION < "2", reason="Only supported for Pydantic v2.")
def test_pydantic_v2_dataclass():
    """Test doc-explained Dataclass support in Pydantic V2."""
    from pydantic import RootModel

    obj = ExampleDataclass(foo="wuz")
    assert to_yaml_str(RootModel[ExampleDataclass](obj)) == "foo: wuz\n"



"""Deprecated types and enums."""

from enum import Enum

from typing_extensions import deprecated  # type: ignore

__all__ = [
    "YamlInt",
    "YamlIntEnum",
    "YamlStr",
    "YamlStrEnum",
]


@deprecated("Use `str` instead.")
class YamlStr(str):
    """YAML-compatible string."""


@deprecated("Use `int` instead.")
class YamlInt(int):
    """YAML-compatible int."""


@deprecated("Use `MyEnum(str, Enum)` instead.")
class YamlStrEnum(str, Enum):
    """String enumeration."""

    def __str__(self) -> str:
        """Return wrapped string."""
        return str.__str__(self)

    def __repr__(self) -> str:
        """Repr string that keeps the class type. You may redefine this."""
        return type(self).__qualname__ + "(" + str.__repr__(self) + ")"


@deprecated("Use `MyEnum(int, Enum)` instead.")
class YamlIntEnum(int, Enum):
    """Integer enumeration."""



"""Models used for testing, just normal `pydantic.BaseModel` objects."""

# mypy: ignore-errors

from typing import Optional

from pydantic import BaseModel, Field
from pydantic.types import SecretBytes, SecretStr

from pydantic_yaml.examples.common import MyIntEnum, MyStrEnum


class Empty(BaseModel):
    """Empty model."""


class A(BaseModel):
    """Class A."""

    a: str


class B(BaseModel):
    """Class B."""

    b: str


class Recursive(BaseModel):
    """Recursive model, which is actually unsupported."""

    inner: Optional["Recursive"]
    a: int  # Doesn't work!


Recursive.update_forward_refs()


class SecretTstModel(BaseModel):
    """Normal model with secret fields. This can't be roundtripped."""

    ss: SecretStr
    sb: SecretBytes


def _encode_secret(obj: SecretStr | SecretBytes | None) -> str | bytes | None:
    """Encode secret value."""
    if obj is None:
        return None
    return obj.get_secret_value()


class SecretTstModelDumpable(SecretTstModel):
    """Round-trippable model. This will save secret fields as the raw values."""

    class Config:
        """Configuration."""

        json_encoders = {SecretStr: _encode_secret, SecretBytes: _encode_secret}


class HasEnums(BaseModel):
    """Base model with enums."""

    opts: MyStrEnum
    vals: list[MyIntEnum]


# pydantic v1


class _Name(BaseModel):
    """First/last names."""

    given: str
    family: str


class UsesRefs(BaseModel):
    """Example for the reference data."""

    bill_to: _Name = Field(alias="bill-to")
    ship_to: _Name = Field(alias="ship-to")

    class Config:
        """Pydantic configuration class."""

        allow_population_by_field_name = True


class CustomRootListStr(BaseModel):
    """Model with a custom root type.

    See Also
    --------
    https://docs.pydantic.dev/usage/models/#custom-root-types
    """

    __root__: list[str]


class CustomRootListObj(BaseModel):
    """Model with a custom root type, list of objects.

    See Also
    --------
    https://docs.pydantic.dev/usage/models/#custom-root-types
    """

    __root__: list[A | B]



"""Gets the version, either installed or dynamically."""

__all__ = ["__version__"]


def __get_version() -> str:
    try:
        from setuptools_scm import get_version  # noqa

        vv = get_version(root="../..", relative_to=__file__)
    except Exception:
        import importlib.metadata as _im

        try:
            vv = _im.version("pydantic_yaml")
        except _im.PackageNotFoundError:
            vv = "0.0.0"
    return vv


__version__ = __get_version()



"""Model for dumping Pydantic models.

See Also
--------
Roundtrip comments with ruamel.yaml
    https://yaml.readthedocs.io/en/latest/detail.html#round-trip-including-comments
    Currently, it's not possible to round-trip comments in `pydantic-yaml`.
    If you need to keep comments, you'll have to have parallel updating and validation.
"""

# mypy: ignore-errors

import json
from io import BytesIO, IOBase, StringIO
from pathlib import Path
from typing import Any, TypeVar
from typing_extensions import Literal  # noqa

from pydantic.version import VERSION as PYDANTIC_VERSION
from ruamel.yaml import YAML

if (PYDANTIC_VERSION < "2") or (PYDANTIC_VERSION > "3"):
    raise ImportError("This module can only be imported in Pydantic v2.")

from pydantic import BaseModel, TypeAdapter
from pydantic.v1 import BaseModel as BaseModelV1
from pydantic.v1 import parse_obj_as

T = TypeVar("T", bound=BaseModel | BaseModelV1)


def _chk_model(model: Any) -> tuple[BaseModel | BaseModelV1, Literal[1, 2]]:
    """Ensure the model passed is a Pydantic model."""
    if isinstance(model, BaseModel):
        return model, 2
    elif isinstance(model, BaseModelV1):
        return model, 1
    raise TypeError(
        "We can currently only write `pydantic.BaseModel` or `pydantic.v1.BaseModel`"
        f"but recieved: {model!r}"
    )


def _write_yaml_model(
    stream: IOBase,
    model: BaseModel | BaseModelV1,
    *,
    default_flow_style: bool | None = None,
    indent: int | None = None,
    map_indent: int | None = None,
    sequence_indent: int | None = None,
    sequence_dash_offset: int | None = None,
    custom_yaml_writer: YAML | None = None,
    **json_kwargs,
) -> None:
    """Write YAML model to the stream object.

    This uses JSON dumping as an intermediary.

    Parameters
    ----------
    stream : IOBase
        The stream to write to.
    model : BaseModel
        The model to write.
    default_flow_style : bool
        Whether to use "flow style" (more human-readable).
        https://yaml.readthedocs.io/en/latest/detail.html?highlight=default_flow_style#indentation-of-block-sequences
    indent : None or int
        General indent value. Leave as None for the default.
    map_indent, sequence_indent, sequence_dash_offset : None or int
        More specific indent values.
    custom_yaml_writer : None or YAML
        An instance of ruamel.yaml.YAML (or a subclass) to use as the writer.
        The above options will be set on it, if given.
    json_kwargs : Any
        Keyword arguments to pass `model.json()`.
    """
    model, vers = _chk_model(model)
    if vers == 1:
        json_val = model.json(**json_kwargs)  # type: ignore
    else:
        json_val = model.model_dump_json(**json_kwargs)  # type: ignore
    val = json.loads(json_val)
    # Allow setting custom writer
    if custom_yaml_writer is None:
        writer = YAML(typ="safe", pure=True)
    elif isinstance(custom_yaml_writer, YAML):
        writer = custom_yaml_writer
    else:
        raise TypeError(f"Please pass a YAML instance or subclass. Got {custom_yaml_writer!r}")
    # Set options
    if default_flow_style is not None:
        writer.default_flow_style = default_flow_style
    writer.indent(mapping=indent, sequence=indent, offset=indent)
    writer.indent(mapping=map_indent, sequence=sequence_indent, offset=sequence_dash_offset)
    # TODO: Configure writer further?
    writer.dump(val, stream)


def to_yaml_str(
    model: BaseModel | BaseModelV1,
    *,
    default_flow_style: bool | None = False,
    indent: int | None = None,
    map_indent: int | None = None,
    sequence_indent: int | None = None,
    sequence_dash_offset: int | None = None,
    custom_yaml_writer: YAML | None = None,
    **json_kwargs,
) -> str:
    """Generate a YAML string representation of the model.

    Parameters
    ----------
    model : BaseModel
        The model to convert.
    default_flow_style : bool
        Whether to use "flow style" (more human-readable).
        https://yaml.readthedocs.io/en/latest/detail.html?highlight=default_flow_style#indentation-of-block-sequences
    indent : None or int
        General indent value. Leave as None for the default.
    map_indent, sequence_indent, sequence_dash_offset : None or int
        More specific indent values.
    custom_yaml_writer : None or YAML
        An instance of ruamel.yaml.YAML (or a subclass) to use as the writer.
        The above options will be set on it, if given.
    json_kwargs : Any
        Keyword arguments to pass `model.json()`.

    Notes
    -----
    This currently uses JSON dumping as an intermediary.
    This means that you can use `json_encoders` in your model.
    """
    model, _ = _chk_model(model)
    stream = StringIO()
    _write_yaml_model(
        stream,
        model,
        default_flow_style=default_flow_style,
        indent=indent,
        map_indent=map_indent,
        sequence_indent=sequence_indent,
        sequence_dash_offset=sequence_dash_offset,
        custom_yaml_writer=custom_yaml_writer,
        **json_kwargs,
    )
    stream.seek(0)
    return stream.read()


def to_yaml_file(
    file: Path | str | IOBase,
    model: BaseModel | BaseModelV1,
    *,
    default_flow_style: bool | None = False,
    indent: int | None = None,
    map_indent: int | None = None,
    sequence_indent: int | None = None,
    sequence_dash_offset: int | None = None,
    custom_yaml_writer: YAML | None = None,
    **json_kwargs,
) -> None:
    """Write a YAML file representation of the model.

    Parameters
    ----------
    file : Path or str or IOBase
        The file path or stream to write to.
    model : BaseModel
        The model to write.
    default_flow_style : bool
        Whether to use "flow style" (more human-readable).
        https://yaml.readthedocs.io/en/latest/detail.html?highlight=default_flow_style#indentation-of-block-sequences
    indent : None or int
        General indent value. Leave as None for the default.
    map_indent, sequence_indent, sequence_dash_offset : None or int
        More specific indent values.
    custom_yaml_writer : None or YAML
        An instance of ruamel.yaml.YAML (or a subclass) to use as the writer.
        The above options will be set on it, if given.
    json_kwargs : Any
        Keyword arguments to pass `model.json()`.

    Notes
    -----
    This currently uses JSON dumping as an intermediary.
    This means that you can use `json_encoders` in your model.
    """
    model, _ = _chk_model(model)
    write_kwargs = dict(
        default_flow_style=default_flow_style,
        indent=indent,
        map_indent=map_indent,
        sequence_indent=sequence_indent,
        sequence_dash_offset=sequence_dash_offset,
        custom_yaml_writer=custom_yaml_writer,
        **json_kwargs,
    )
    if isinstance(file, IOBase):  # open file handle
        _write_yaml_model(file, model, **write_kwargs)
        return

    if isinstance(file, str):  # local path to file
        file = Path(file).resolve()
    elif isinstance(file, Path):
        file = file.resolve()
    else:
        raise TypeError(f"Expected Path, str, or stream, but got {file!r}")

    with file.open(mode="w") as f:
        _write_yaml_model(f, model, **write_kwargs)
        return


def parse_yaml_raw_as(model_type: type[T], raw: str | bytes | IOBase) -> T:
    """Parse raw YAML string as the passed model type.

    Parameters
    ----------
    model_type : Type[BaseModel]
        The resulting model type.
    raw : str or bytes or IOBase
        The YAML string or stream.
    """
    stream: IOBase
    if isinstance(raw, str):
        stream = StringIO(raw)
    elif isinstance(raw, bytes):
        stream = BytesIO(raw)
    elif isinstance(raw, IOBase):
        stream = raw
    else:
        raise TypeError(f"Expected str, bytes or IO, but got {raw!r}")
    reader = YAML(typ="safe", pure=True)  # YAML 1.2 support
    objects = reader.load(stream)
    if isinstance(model_type, type) and issubclass(model_type, BaseModelV1):
        return parse_obj_as(model_type, objects)  # type:ignore
    else:
        ta = TypeAdapter(model_type)  # type: ignore
        return ta.validate_python(objects)


def parse_yaml_file_as(model_type: type[T], file: Path | str | IOBase) -> T:
    """Parse YAML file as the passed model type.

    Parameters
    ----------
    model_type : Type[BaseModel]
        The resulting model type.
    file : Path or str or IOBase
        The file path or stream to read from.
    """
    # Short-circuit
    if isinstance(file, IOBase):
        return parse_yaml_raw_as(model_type, raw=file)

    if isinstance(file, str):
        file = Path(file).resolve()
    elif isinstance(file, Path):
        file = file.resolve()
    else:
        raise TypeError(f"Expected Path, str or IO, but got {file!r}")

    with file.open(mode="r") as f:
        return parse_yaml_raw_as(model_type, f)



"""Common types used in examples."""

from enum import Enum


class MyStrEnum(str, Enum):
    """String enumeration for testing."""

    option1 = "option1"
    option2 = "option2"


class MyIntEnum(int, Enum):
    """Integer enumeration for testing."""

    v1 = 1
    v2 = 2



"""Bad behaviors that should cause errors."""

from typing import Any

import pytest
from pydantic import BaseModel

from pydantic_yaml import parse_yaml_file_as, parse_yaml_raw_as, to_yaml_str
from pydantic_yaml.examples.base_models import A, B, Recursive, root


def test_no_load_recursive():
    """Test properly rejecting loading a recursive model."""
    with pytest.raises(ValueError):
        parse_yaml_file_as(Recursive, root / "recursive.yaml")


@pytest.mark.parametrize("obj", ["foo", {}, set(), ..., ["some-object", 1]])
def test_no_dump(obj: Any):
    """Ensure that we don't support dumping these objects."""
    with pytest.raises((ValueError, TypeError)):
        to_yaml_str(obj)


@pytest.mark.parametrize(
    ["model_type", "raw"],
    [
        (B, None),
        (A, 3),
        (B, "aaa"),
        (A, "aaaaaaaaaaa"),
    ],
)
def test_no_load(model_type: type[BaseModel], raw: Any):
    """Ensure that we don't support loading these objects."""
    with pytest.raises((ValueError, TypeError)):
        parse_yaml_raw_as(model_type, raw)



"""Module for loading Pydantic models."""

__all__ = ["parse_yaml_raw_as", "parse_yaml_file_as"]

import warnings
from pydantic.version import VERSION as PYDANTIC_VERSION

if (PYDANTIC_VERSION > "1") and (PYDANTIC_VERSION < "2"):
    from pydantic_yaml._internals.v1 import parse_yaml_raw_as, parse_yaml_file_as
elif (PYDANTIC_VERSION > "2") and (PYDANTIC_VERSION < "3"):
    from pydantic_yaml._internals.v2 import parse_yaml_raw_as, parse_yaml_file_as
else:
    raise ImportError(f"Unsupported version of Pydantic: {PYDANTIC_VERSION!r}")

warnings.warn(DeprecationWarning("This module is deprecated; use `from pydantic_yaml` instead."))



"""Regression tests for when 'type' to parse as is not a class."""

import sys
from typing import Union

from pydantic_yaml import parse_yaml_raw_as
from pydantic import BaseModel


class MyStrModel(BaseModel):
    """String model."""

    my_prop: str


class MyListModel(BaseModel):
    """List of strings model."""

    my_prop: list[str]


def test_non_class_types_union() -> None:
    """Test non-class types."""
    inst1: MyStrModel | MyListModel = parse_yaml_raw_as(
        Union[MyStrModel, MyListModel],  # type: ignore  # noqa
        "my_prop: my_yaml_string",
    )
    assert isinstance(inst1, MyStrModel)
    inst2: MyStrModel | MyListModel = parse_yaml_raw_as(
        Union[MyStrModel, MyListModel],  # type: ignore  # noqa
        "my_prop: [my_yaml_string]",
    )
    assert isinstance(inst2, MyListModel)


if sys.version_info > (3, 10, 0):

    def test_non_class_types_uniontype() -> None:
        """Test non-class types."""
        inst1: MyStrModel | MyListModel = parse_yaml_raw_as(  # type: ignore
            MyStrModel | MyListModel,  # type: ignore
            "my_prop: my_yaml_string",
        )
        assert isinstance(inst1, MyStrModel)
        inst2: MyStrModel | MyListModel = parse_yaml_raw_as(  # type: ignore
            MyStrModel | MyListModel,  # type: ignore
            "my_prop: [my_yaml_string]",
        )
        assert isinstance(inst2, MyListModel)



"""Tests for Pydantic v2-v1 compatibility."""

import pytest

from pydantic.version import VERSION as PYDANTIC_VERSION
from pydantic_yaml import to_yaml_str, parse_yaml_raw_as


@pytest.mark.skipif(PYDANTIC_VERSION < "2", reason="Only supported for Pydantic v2.")
def test_pydantic_v2_v1_compat() -> None:
    """Test v1-compatibility in pydantic v2."""
    from pydantic.v1 import BaseModel

    class MyModel(BaseModel):
        """Simple model for dumping."""

        v: str = "vee"

    yml = to_yaml_str(MyModel(v="tree"))  # type: ignore
    obj = parse_yaml_raw_as(MyModel, yml)  # type: ignore
    assert obj.v == "tree"



"""Support for Pydantic v1 interface."""

# mypy: ignore-errors

import json
from io import StringIO, BytesIO, IOBase
from pathlib import Path
from typing import Any, TypeVar

from pydantic.version import VERSION as PYDANTIC_VERSION
from ruamel.yaml import YAML

if PYDANTIC_VERSION > "2":
    raise ImportError("This module can only be imported in Pydantic v1.")

from pydantic import BaseModel, parse_obj_as

T = TypeVar("T", bound=BaseModel)


def _chk_model(model: Any) -> BaseModel:
    """Ensure the model passed is a Pydantic model."""
    if isinstance(model, BaseModel):
        return model
    raise TypeError("We can currently only write `pydantic.BaseModel`, " f"but recieved: {model!r}")


def parse_yaml_raw_as(model_type: type[T], raw: str | bytes | IOBase) -> T:
    """Parse raw YAML string as the passed model type.

    Parameters
    ----------
    model_type : Type[BaseModel]
        The resulting model type.
    raw : str or bytes or IOBase
        The YAML string or stream.
    """
    stream: IOBase
    if isinstance(raw, str):
        stream = StringIO(raw)
    elif isinstance(raw, bytes):
        stream = BytesIO(raw)
    elif isinstance(raw, IOBase):
        stream = raw
    else:
        raise TypeError(f"Expected str, bytes or IO, but got {raw!r}")
    reader = YAML(typ="safe", pure=True)  # YAML 1.2 support
    objects = reader.load(stream)
    return parse_obj_as(model_type, objects)  # type:ignore


def parse_yaml_file_as(model_type: type[T], file: Path | str | IOBase) -> T:
    """Parse YAML file as the passed model type.

    Parameters
    ----------
    model_type : Type[BaseModel]
        The resulting model type.
    file : Path or str or IOBase
        The file path or stream to read from.
    """
    # Short-circuit
    if isinstance(file, IOBase):
        return parse_yaml_raw_as(model_type, raw=file)

    if isinstance(file, str):
        file = Path(file).resolve()
    elif isinstance(file, Path):
        file = file.resolve()
    else:
        raise TypeError(f"Expected Path, str or IO, but got {file!r}")

    with file.open(mode="r") as f:
        return parse_yaml_raw_as(model_type, f)


def _write_yaml_model(
    stream: IOBase,
    model: BaseModel,
    *,
    default_flow_style: bool | None = None,
    indent: int | None = None,
    map_indent: int | None = None,
    sequence_indent: int | None = None,
    sequence_dash_offset: int | None = None,
    custom_yaml_writer: YAML | None = None,
    **json_kwargs,
) -> None:
    """Write YAML model to the stream object.

    This uses JSON dumping as an intermediary.

    Parameters
    ----------
    stream : IOBase
        The stream to write to.
    model : BaseModel
        The model to write.
    default_flow_style : bool
        Whether to use "flow style" (more human-readable).
        https://yaml.readthedocs.io/en/latest/detail.html?highlight=default_flow_style#indentation-of-block-sequences
    indent : None or int
        General indent value. Leave as None for the default.
    map_indent, sequence_indent, sequence_dash_offset : None or int
        More specific indent values.
    custom_yaml_writer : None or YAML
        An instance of ruamel.yaml.YAML (or a subclass) to use as the writer.
        The above options will be set on it, if given.
    json_kwargs : Any
        Keyword arguments to pass `model.json()`.
    """
    model = _chk_model(model)
    json_val = model.json(**json_kwargs)  # type: ignore
    val = json.loads(json_val)
    # Allow setting custom writer
    if custom_yaml_writer is None:
        writer = YAML(typ="safe", pure=True)
    elif isinstance(custom_yaml_writer, YAML):
        writer = custom_yaml_writer
    else:
        raise TypeError(f"Please pass a YAML instance or subclass. Got {custom_yaml_writer!r}")
    # Set options
    if default_flow_style is not None:
        writer.default_flow_style = default_flow_style
    writer.indent(mapping=indent, sequence=indent, offset=indent)
    writer.indent(mapping=map_indent, sequence=sequence_indent, offset=sequence_dash_offset)
    # TODO: Configure writer further?
    writer.dump(val, stream)


def to_yaml_str(
    model: BaseModel,
    *,
    default_flow_style: bool | None = False,
    indent: int | None = None,
    map_indent: int | None = None,
    sequence_indent: int | None = None,
    sequence_dash_offset: int | None = None,
    custom_yaml_writer: YAML | None = None,
    **json_kwargs,
) -> str:
    """Generate a YAML string representation of the model.

    Parameters
    ----------
    model : BaseModel
        The model to convert.
    default_flow_style : bool
        Whether to use "flow style" (more human-readable).
        https://yaml.readthedocs.io/en/latest/detail.html?highlight=default_flow_style#indentation-of-block-sequences
    indent : None or int
        General indent value. Leave as None for the default.
    map_indent, sequence_indent, sequence_dash_offset : None or int
        More specific indent values.
    custom_yaml_writer : None or YAML
        An instance of ruamel.yaml.YAML (or a subclass) to use as the writer.
        The above options will be set on it, if given.
    json_kwargs : Any
        Keyword arguments to pass `model.json()`.

    Notes
    -----
    This currently uses JSON dumping as an intermediary.
    This means that you can use `json_encoders` in your model.
    """
    model = _chk_model(model)
    stream = StringIO()
    _write_yaml_model(
        stream,
        model,
        default_flow_style=default_flow_style,
        indent=indent,
        map_indent=map_indent,
        sequence_indent=sequence_indent,
        sequence_dash_offset=sequence_dash_offset,
        custom_yaml_writer=custom_yaml_writer,
        **json_kwargs,
    )
    stream.seek(0)
    return stream.read()


def to_yaml_file(
    file: Path | str | IOBase,
    model: BaseModel,
    *,
    default_flow_style: bool | None = False,
    indent: int | None = None,
    map_indent: int | None = None,
    sequence_indent: int | None = None,
    sequence_dash_offset: int | None = None,
    custom_yaml_writer: YAML | None = None,
    **json_kwargs,
) -> None:
    """Write a YAML file representation of the model.

    Parameters
    ----------
    file : Path or str or IOBase
        The file path or stream to write to.
    model : BaseModel
        The model to write.
    default_flow_style : bool
        Whether to use "flow style" (more human-readable).
        https://yaml.readthedocs.io/en/latest/detail.html?highlight=default_flow_style#indentation-of-block-sequences
    indent : None or int
        General indent value. Leave as None for the default.
    map_indent, sequence_indent, sequence_dash_offset : None or int
        More specific indent values.
    custom_yaml_writer : None or YAML
        An instance of ruamel.yaml.YAML (or a subclass) to use as the writer.
        The above options will be set on it, if given.
    json_kwargs : Any
        Keyword arguments to pass `model.json()`.

    Notes
    -----
    This currently uses JSON dumping as an intermediary.
    This means that you can use `json_encoders` in your model.
    """
    model = _chk_model(model)
    write_kwargs = dict(
        default_flow_style=default_flow_style,
        indent=indent,
        map_indent=map_indent,
        sequence_indent=sequence_indent,
        sequence_dash_offset=sequence_dash_offset,
        custom_yaml_writer=custom_yaml_writer,
        **json_kwargs,
    )
    if isinstance(file, IOBase):  # open file handle
        _write_yaml_model(file, model, **write_kwargs)
        return

    if isinstance(file, str):  # local path to file
        file = Path(file).resolve()
    elif isinstance(file, Path):
        file = file.resolve()
    else:
        raise TypeError(f"Expected Path, str, or stream, but got {file!r}")

    with file.open(mode="w") as f:
        _write_yaml_model(f, model, **write_kwargs)
        return



"""Basic tests for the README examples."""

import pytest
import pydantic


@pytest.mark.skipif(pydantic.version.VERSION >= "2", reason="Example is for Pydantic V1")
def test_readme_basic_usage_v1():
    """Basic usage example from README."""
    from enum import Enum

    from pydantic import BaseModel, validator

    from pydantic_yaml import parse_yaml_raw_as, to_yaml_str

    class MyEnum(str, Enum):
        """A custom enumeration that is YAML-safe."""

        a = "a"
        b = "b"

    class InnerModel(BaseModel):
        """A normal pydantic model that can be used as an inner class."""

        fld: float = 1.0

    class MyModel(BaseModel):
        """Our custom class, with a `.yaml()` method.

        The `parse_raw()` and `parse_file()` methods are also updated to be able to
        handle `content_type='application/yaml'`, `protocol="yaml"` and file names
        ending with `.yml`/`.yaml`
        """

        x: int = 1
        e: MyEnum = MyEnum.a
        m: InnerModel = InnerModel()

        @validator("x")
        def _chk_x(cls, v: int) -> int:  # noqa
            """You can add your normal pydantic validators, like this one."""
            assert v > 0
            return v

    m1 = MyModel(x=2, e="b", m=InnerModel(fld=1.5))  # type: ignore

    # This dumps to YAML and JSON respectively
    yml = to_yaml_str(m1)
    jsn = m1.json()

    # This parses YAML as the MyModel type
    m2 = parse_yaml_raw_as(MyModel, yml)
    assert m1 == m2

    # JSON is also valid YAML, so this works too
    m3 = parse_yaml_raw_as(MyModel, jsn)
    assert m1 == m3


@pytest.mark.skipif(pydantic.version.VERSION < "2", reason="Example is for Pydantic V2")
def test_readme_basic_usage_v2():
    """Basic usage example from README."""
    from enum import Enum

    from pydantic import BaseModel, field_validator

    from pydantic_yaml import parse_yaml_raw_as, to_yaml_str

    class MyEnum(str, Enum):
        """A custom enumeration that is YAML-safe."""

        a = "a"
        b = "b"

    class InnerModel(BaseModel):
        """A normal pydantic model that can be used as an inner class."""

        fld: float = 1.0

    class MyModel(BaseModel):
        """Our custom class, with a `.yaml()` method.

        The `parse_raw()` and `parse_file()` methods are also updated to be able to
        handle `content_type='application/yaml'`, `protocol="yaml"` and file names
        ending with `.yml`/`.yaml`
        """

        x: int = 1
        e: MyEnum = MyEnum.a
        m: InnerModel = InnerModel()

        @field_validator("x")
        def _chk_x(cls, v: int) -> int:  # noqa
            """You can add your normal pydantic validators, like this one."""
            assert v > 0
            return v

    m1 = MyModel(x=2, e="b", m=InnerModel(fld=1.5))  # type: ignore

    # This dumps to YAML and JSON respectively
    yml = to_yaml_str(m1)
    jsn = m1.model_dump_json()

    # This parses YAML as the MyModel type
    m2 = parse_yaml_raw_as(MyModel, yml)
    assert m1 == m2

    # JSON is also valid YAML, so this works too
    m3 = parse_yaml_raw_as(MyModel, jsn)
    assert m1 == m3
